"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createLanguageService = exports.getDocumentLanguageService = void 0;
const vscode_languageserver_textdocument_1 = require("vscode-languageserver-textdocument");
const shared = require("@volar/shared");
const sourceFile_1 = require("./sourceFile");
const localTypes = require("./utils/localTypes");
const upath = require("upath");
const tsPluginApis = require("./tsPluginApis");
const tsProgramApis = require("./tsProgramApis");
// vue services
const completions = require("./services/completion");
const completionResolve = require("./services/completionResolve");
const autoCreateQuotes = require("./services/autoCreateQuotes");
const autoClosingTags = require("./services/autoClosingTags");
const autoWrapBrackets = require("./services/autoWrapParentheses");
const refAutoClose = require("./services/refAutoClose");
const hover = require("./services/hover");
const diagnostics = require("./services/diagnostics");
const formatting = require("./services/formatting");
const definitions = require("./services/definition");
const references = require("./services/references");
const rename = require("./services/rename");
const codeActions = require("./services/codeAction");
const codeActionResolve = require("./services/codeActionResolve");
const documentHighlight = require("./services/documentHighlight");
const documentSymbol = require("./services/documentSymbol");
const documentLink = require("./services/documentLinks");
const documentColor = require("./services/documentColor");
const selectionRanges = require("./services/selectionRanges");
const signatureHelp = require("./services/signatureHelp");
const colorPresentations = require("./services/colorPresentation");
const semanticTokens = require("./services/semanticTokens");
const foldingRanges = require("./services/foldingRanges");
const codeLens = require("./services/referencesCodeLens");
const codeLensResolve = require("./services/referencesCodeLensResolve");
const callHierarchy = require("./services/callHierarchy");
const linkedEditingRanges = require("./services/linkedEditingRange");
const tagNameCase = require("./services/tagNameCase");
const workspaceSymbol = require("./services/workspaceSymbol");
const d3 = require("./services/d3");
// context
const fs = require("fs");
const emmet = require("@vscode/emmet-helper");
const css = require("vscode-css-languageservice");
const html = require("vscode-html-languageservice");
const json = require("vscode-json-languageservice");
const ts2 = require("vscode-typescript-languageservice");
const pug = require("vscode-pug-languageservice");
const sourceFiles_1 = require("./sourceFiles");
function getDocumentLanguageService(modules, getPreferences, getFormatOptions, formatters) {
    const htmlDocuments = new WeakMap();
    const vueDocuments = new WeakMap();
    const services = createServices(modules.typescript);
    const context = {
        compilerOptions: {},
        modules: {
            typescript: modules.typescript,
            emmet,
            css,
            html,
            json,
            ts: ts2,
            pug
        },
        ...services,
        getHtmlDocument,
        getVueDocument,
    };
    return {
        doFormatting: formatting.register(context, getPreferences, getFormatOptions, formatters),
        getFoldingRanges: foldingRanges.register(context, getPreferences, getFormatOptions),
        getSelectionRanges: selectionRanges.register(context, getPreferences, getFormatOptions),
        doQuoteComplete: autoCreateQuotes.register(context),
        doTagComplete: autoClosingTags.register(context),
        doParentheseWrap: autoWrapBrackets.register(context),
        findLinkedEditingRanges: linkedEditingRanges.register(context),
        findDocumentSymbols: documentSymbol.register(context, getPreferences, getFormatOptions),
        findDocumentColors: documentColor.register(context),
        getColorPresentations: colorPresentations.register(context),
    };
    function getVueDocument(document) {
        if (document.languageId !== 'vue')
            return;
        const cacheVueDoc = vueDocuments.get(document);
        if (cacheVueDoc) {
            const oldText = cacheVueDoc.getTextDocument().getText();
            const newText = document.getText();
            if (oldText.length !== newText.length || oldText !== newText) {
                cacheVueDoc.update(document.getText(), document.version.toString());
            }
            return cacheVueDoc;
        }
        const vueDoc = (0, sourceFile_1.createSourceFile)(document.uri, document.getText(), document.version.toString(), context);
        vueDocuments.set(document, vueDoc);
        return vueDoc;
    }
    function getHtmlDocument(document) {
        if (document.languageId !== 'vue')
            return;
        const cache = htmlDocuments.get(document);
        if (cache) {
            const [cacheVersion, cacheHtmlDoc] = cache;
            if (cacheVersion === document.version) {
                return cacheHtmlDoc;
            }
        }
        const htmlDoc = context.htmlLs.parseHTMLDocument(document);
        htmlDocuments.set(document, [document.version, htmlDoc]);
        return htmlDoc;
    }
}
exports.getDocumentLanguageService = getDocumentLanguageService;
function createLanguageService(modules, vueHost, isTsPlugin = false) {
    var _a, _b, _c;
    const { typescript: ts } = modules;
    const isVue2 = ((_a = vueHost.getVueCompilationSettings) === null || _a === void 0 ? void 0 : _a.call(vueHost).experimentalCompatMode) === 2;
    let vueProjectVersion;
    let scriptContentVersion = 0; // only update by `<script>` / `<script setup>` / *.ts content
    let scriptProjectVersion = 0; // update by script LS virtual files / *.ts
    let templateProjectVersion = 0;
    let lastScriptProjectVersionWhenTemplateProjectVersionUpdate = -1;
    const documents = shared.createPathMap();
    const sourceFiles = (0, sourceFiles_1.createSourceFiles)();
    const templateScriptUpdateUris = new Set();
    const initProgressCallback = [];
    const blockingRequests = new Set();
    const templateTsHost = createTsLsHost('template');
    const scriptTsHost = createTsLsHost('script');
    const templateTsLsRaw = ts.createLanguageService(templateTsHost);
    const scriptTsLsRaw = ts.createLanguageService(scriptTsHost);
    const templateTsLs = ts2.createLanguageService(ts, templateTsHost, templateTsLsRaw);
    const scriptTsLs = ts2.createLanguageService(ts, scriptTsHost, scriptTsLsRaw);
    const localTypesScript = ts.ScriptSnapshot.fromString(localTypes.getTypesCode(isVue2));
    const compilerHost = ts.createCompilerHost(vueHost.getCompilationSettings());
    const documentContext = {
        resolveReference(ref, base) {
            var _a, _b;
            const isUri = base.indexOf('://') >= 0;
            const resolveResult = ts.resolveModuleName(ref, isUri ? shared.uriToFsPath(base) : base, vueHost.getCompilationSettings(), compilerHost);
            const failedLookupLocations = resolveResult.failedLookupLocations;
            const dirs = new Set();
            const fileExists = (_a = vueHost.fileExists) !== null && _a !== void 0 ? _a : ts.sys.fileExists;
            const directoryExists = (_b = vueHost.directoryExists) !== null && _b !== void 0 ? _b : ts.sys.directoryExists;
            for (const failed of failedLookupLocations) {
                let path = failed;
                const fileName = upath.basename(path);
                if (fileName === 'index.d.ts' || fileName === '*.d.ts') {
                    dirs.add(upath.dirname(path));
                }
                if (path.endsWith('.d.ts')) {
                    path = upath.removeExt(upath.removeExt(path, '.ts'), '.d');
                }
                else {
                    continue;
                }
                if (fileExists(path)) {
                    return isUri ? shared.fsPathToUri(path) : path;
                }
            }
            for (const dir of dirs) {
                if (directoryExists(dir)) {
                    return isUri ? shared.fsPathToUri(dir) : dir;
                }
            }
            return undefined;
        },
    };
    const services = createServices(modules.typescript, vueHost);
    const context = {
        compilerOptions: (_c = (_b = vueHost.getVueCompilationSettings) === null || _b === void 0 ? void 0 : _b.call(vueHost)) !== null && _c !== void 0 ? _c : {},
        modules: {
            typescript: modules.typescript,
            emmet,
            css,
            html,
            json,
            ts: ts2,
            pug
        },
        ...services,
        vueHost,
        sourceFiles,
        templateTsHost,
        scriptTsHost,
        templateTsLsRaw,
        scriptTsLsRaw,
        templateTsLs,
        scriptTsLs,
        documentContext,
        getTsLs: (lsType) => lsType === 'template' ? templateTsLs : scriptTsLs,
        getTextDocument: getHostDocument,
    };
    const _callHierarchy = callHierarchy.register(context);
    const findDefinition = definitions.register(context);
    const renames = rename.register(context);
    let tsPluginProxy;
    let tsProgramProxy;
    ;
    return {
        doValidation: publicApiHook(diagnostics.register(context, () => update(true)), false, false),
        findDefinition: publicApiHook(findDefinition.on, isTemplateScriptPosition),
        findReferences: publicApiHook(references.register(context), true),
        findTypeDefinition: publicApiHook(findDefinition.onType, isTemplateScriptPosition),
        callHierarchy: {
            doPrepare: publicApiHook(_callHierarchy.doPrepare, isTemplateScriptPosition),
            getIncomingCalls: publicApiHook(_callHierarchy.getIncomingCalls, true),
            getOutgoingCalls: publicApiHook(_callHierarchy.getOutgoingCalls, true),
        },
        prepareRename: publicApiHook(renames.prepareRename, isTemplateScriptPosition),
        doRename: publicApiHook(renames.doRename, true),
        getEditsForFileRename: publicApiHook(renames.onRenameFile, false),
        getSemanticTokens: publicApiHook(semanticTokens.register(context, () => update(true)), false),
        doHover: publicApiHook(hover.register(context), isTemplateScriptPosition),
        doComplete: publicApiHook(completions.register(context, () => scriptContentVersion), isTemplateScriptPosition),
        getCodeActions: publicApiHook(codeActions.register(context), false),
        doCodeActionResolve: publicApiHook(codeActionResolve.register(context), false),
        doCompletionResolve: publicApiHook(completionResolve.register(context), false),
        doReferencesCodeLensResolve: publicApiHook(codeLensResolve.register(context), false),
        getSignatureHelp: publicApiHook(signatureHelp.register(context), false),
        getReferencesCodeLens: publicApiHook(codeLens.register(context), false),
        findDocumentHighlights: publicApiHook(documentHighlight.register(context), false),
        findDocumentLinks: publicApiHook(documentLink.register(context), false),
        findWorkspaceSymbols: publicApiHook(workspaceSymbol.register(context), false),
        dispose: () => {
            scriptTsLs.dispose();
            templateTsLs.dispose();
        },
        updateHtmlCustomData: services.updateHtmlCustomData,
        updateCssCustomData: services.updateCssCustomData,
        __internal__: {
            rootPath: vueHost.getCurrentDirectory(),
            get tsPlugin() {
                if (!tsPluginProxy) {
                    tsPluginProxy = createTsPluginProxy();
                }
                return tsPluginProxy;
            },
            get tsProgramProxy() {
                if (!tsProgramProxy) {
                    tsProgramProxy = createTsProgramProxy();
                }
                return tsProgramProxy;
            },
            context,
            onInitProgress(cb) {
                initProgressCallback.push(cb);
            },
            getLocalTypesFiles: (lsType) => {
                const fileNames = getLocalTypesFiles(lsType);
                const code = localTypes.getTypesCode(isVue2);
                return {
                    fileNames,
                    code,
                };
            },
            getContext: publicApiHook(() => context),
            getD3: publicApiHook(d3.register(context)),
            detectTagNameCase: publicApiHook(tagNameCase.register(context)),
            doRefAutoClose: publicApiHook(refAutoClose.register(context), false),
        },
    };
    function getLocalTypesFiles(lsType) {
        if (lsType === 'script')
            return [];
        return sourceFiles.getDirs().map(dir => upath.join(dir, localTypes.typesFileName));
    }
    function createTsPluginProxy() {
        // ts plugin proxy
        const _tsPluginApis = tsPluginApis.register(context);
        const tsPlugin = {
            getSemanticDiagnostics: apiHook(scriptTsLsRaw.getSemanticDiagnostics, false),
            getEncodedSemanticClassifications: apiHook(scriptTsLsRaw.getEncodedSemanticClassifications, false),
            getCompletionsAtPosition: apiHook(_tsPluginApis.getCompletionsAtPosition, false),
            getCompletionEntryDetails: apiHook(scriptTsLsRaw.getCompletionEntryDetails, false),
            getCompletionEntrySymbol: apiHook(scriptTsLsRaw.getCompletionEntrySymbol, false),
            getQuickInfoAtPosition: apiHook(scriptTsLsRaw.getQuickInfoAtPosition, false),
            getSignatureHelpItems: apiHook(scriptTsLsRaw.getSignatureHelpItems, false),
            getRenameInfo: apiHook(scriptTsLsRaw.getRenameInfo, false),
            findRenameLocations: apiHook(_tsPluginApis.findRenameLocations, true),
            getDefinitionAtPosition: apiHook(_tsPluginApis.getDefinitionAtPosition, false),
            getDefinitionAndBoundSpan: apiHook(_tsPluginApis.getDefinitionAndBoundSpan, false),
            getTypeDefinitionAtPosition: apiHook(_tsPluginApis.getTypeDefinitionAtPosition, false),
            getImplementationAtPosition: apiHook(_tsPluginApis.getImplementationAtPosition, false),
            getReferencesAtPosition: apiHook(_tsPluginApis.getReferencesAtPosition, true),
            findReferences: apiHook(_tsPluginApis.findReferences, true),
            // TODO: now is handle by vue server
            // prepareCallHierarchy: apiHook(tsLanguageService.rawLs.prepareCallHierarchy, false),
            // provideCallHierarchyIncomingCalls: apiHook(tsLanguageService.rawLs.provideCallHierarchyIncomingCalls, false),
            // provideCallHierarchyOutgoingCalls: apiHook(tsLanguageService.rawLs.provideCallHierarchyOutgoingCalls, false),
            // getEditsForFileRename: apiHook(tsLanguageService.rawLs.getEditsForFileRename, false),
            // TODO
            // getCodeFixesAtPosition: apiHook(tsLanguageService.rawLs.getCodeFixesAtPosition, false),
            // getCombinedCodeFix: apiHook(tsLanguageService.rawLs.getCombinedCodeFix, false),
            // applyCodeActionCommand: apiHook(tsLanguageService.rawLs.applyCodeActionCommand, false),
            // getApplicableRefactors: apiHook(tsLanguageService.rawLs.getApplicableRefactors, false),
            // getEditsForRefactor: apiHook(tsLanguageService.rawLs.getEditsForRefactor, false),
        };
        return tsPlugin;
    }
    function createTsProgramProxy() {
        // ts program proxy
        const tsProgram = scriptTsLsRaw.getProgram(); // TODO: handle template ls?
        if (!tsProgram)
            throw '!tsProgram';
        const tsProgramApis_2 = tsProgramApis.register(context);
        const tsProgramApis_3 = {
            emit: apiHook(tsProgramApis_2.emit),
            getRootFileNames: apiHook(tsProgramApis_2.getRootFileNames),
            getSemanticDiagnostics: apiHook(tsProgramApis_2.getSemanticDiagnostics),
            getSyntacticDiagnostics: apiHook(tsProgramApis_2.getSyntacticDiagnostics),
            getGlobalDiagnostics: apiHook(tsProgramApis_2.getGlobalDiagnostics),
        };
        const tsProgramProxy = new Proxy(tsProgram, {
            get: (target, property) => {
                return tsProgramApis_3[property] || target[property];
            },
        });
        return tsProgramProxy;
    }
    function isTemplateScriptPosition(uri, pos) {
        const sourceFile = sourceFiles.get(uri);
        if (!sourceFile) {
            return false;
        }
        for (const sourceMap of sourceFile.getTsSourceMaps()) {
            if (sourceMap.lsType === 'script')
                continue;
            for (const _ of sourceMap.getMappedRanges(pos, pos, data => data.vueTag === 'template'
                || data.vueTag === 'style' // handle CSS variable injection to fix https://github.com/johnsoncodehk/volar/issues/777
            )) {
                return true;
            }
        }
        for (const sourceMap of sourceFile.getHtmlSourceMaps()) {
            for (const _ of sourceMap.getMappedRanges(pos)) {
                return true;
            }
        }
        return false;
    }
    function apiHook(api, shouldUpdateTemplateScript = true) {
        const handler = {
            apply(target, thisArg, argumentsList) {
                const _shouldUpdateTemplateScript = typeof shouldUpdateTemplateScript === 'boolean' ? shouldUpdateTemplateScript : shouldUpdateTemplateScript.apply(null, argumentsList);
                update(_shouldUpdateTemplateScript);
                return target.apply(thisArg, argumentsList);
            }
        };
        return new Proxy(api, handler);
    }
    function publicApiHook(api, shouldUpdateTemplateScript = true, blockNewRequest = true) {
        const handler = {
            async apply(target, thisArg, argumentsList) {
                for (const runningRequest of blockingRequests) {
                    await runningRequest;
                }
                const _shouldUpdateTemplateScript = typeof shouldUpdateTemplateScript === 'boolean' ? shouldUpdateTemplateScript : shouldUpdateTemplateScript.apply(null, argumentsList);
                update(_shouldUpdateTemplateScript);
                const runner = target.apply(thisArg, argumentsList);
                if (blockNewRequest && runner instanceof Promise) {
                    blockingRequests.add(runner);
                    runner.then(() => blockingRequests.delete(runner));
                }
                return runner;
            }
        };
        return new Proxy(api, handler);
    }
    function update(shouldUpdateTemplateScript) {
        var _a, _b;
        const newVueProjectVersion = (_a = vueHost.getVueProjectVersion) === null || _a === void 0 ? void 0 : _a.call(vueHost);
        if (newVueProjectVersion === undefined || newVueProjectVersion !== vueProjectVersion) {
            vueProjectVersion = newVueProjectVersion;
            const newFileUris = new Set([...vueHost.getScriptFileNames()].filter(file => file.endsWith('.vue')).map(shared.fsPathToUri));
            const removeUris = [];
            const addUris = [];
            const updateUris = [];
            for (const sourceFile of sourceFiles.getAll()) {
                const fileName = shared.uriToFsPath(sourceFile.uri);
                if (!newFileUris.has(sourceFile.uri) && !((_b = vueHost.fileExists) === null || _b === void 0 ? void 0 : _b.call(vueHost, fileName))) {
                    // delete
                    removeUris.push(sourceFile.uri);
                }
                else {
                    // update
                    const newVersion = vueHost.getScriptVersion(fileName);
                    if (sourceFile.getVersion() !== newVersion) {
                        updateUris.push(sourceFile.uri);
                    }
                }
            }
            for (const newUri of newFileUris) {
                if (!sourceFiles.get(newUri)) {
                    // add
                    addUris.push(newUri);
                }
            }
            // if (tsFileChanged) {
            // 	scriptContentVersion++;
            // 	scriptProjectVersion++;
            // 	templateProjectVersion++;
            // 	// TODO: template global properties can't update by .d.ts definition
            // 	// wait for https://github.com/johnsoncodehk/volar/issues/455
            // 	// updates.length = 0;
            // 	// for (const fileName of oldFiles) {
            // 	// 	if (newFiles.has(fileName)) {
            // 	// 		if (fileName.endsWith('.vue')) {
            // 	// 			updates.push(fileName);
            // 	// 		}
            // 	// 	}
            // 	// }
            // }
            const finalUpdateUris = addUris.concat(updateUris);
            if (removeUris.length) {
                unsetSourceFiles(removeUris);
            }
            if (finalUpdateUris.length) {
                updateSourceFiles(finalUpdateUris, shouldUpdateTemplateScript);
            }
        }
        else if (shouldUpdateTemplateScript && templateScriptUpdateUris.size) {
            updateSourceFiles([], shouldUpdateTemplateScript);
        }
    }
    function createTsLsHost(lsType) {
        const scriptSnapshots = new Map();
        const tsHost = {
            ...vueHost,
            fileExists: vueHost.fileExists
                ? fileName => {
                    var _a, _b;
                    // .vue.js -> .vue
                    // .vue.ts -> .vue
                    // .vue.d.ts (never)
                    const fileNameTrim = upath.trimExt(fileName);
                    if (fileNameTrim.endsWith('.vue')) {
                        const uri = shared.fsPathToUri(fileNameTrim);
                        const sourceFile = sourceFiles.get(uri);
                        if (!sourceFile) {
                            const fileExists = !!((_a = vueHost.fileExists) === null || _a === void 0 ? void 0 : _a.call(vueHost, fileNameTrim));
                            if (fileExists) {
                                updateSourceFiles([uri], false); // create virtual files
                            }
                        }
                        return sourceFiles.getTsDocuments(lsType).has(shared.fsPathToUri(fileName));
                    }
                    else {
                        return !!((_b = vueHost.fileExists) === null || _b === void 0 ? void 0 : _b.call(vueHost, fileName));
                    }
                }
                : undefined,
            getProjectVersion: () => {
                var _a;
                return ((_a = vueHost.getProjectVersion) === null || _a === void 0 ? void 0 : _a.call(vueHost)) + '-' + (lsType === 'template' ? templateProjectVersion : scriptProjectVersion).toString();
            },
            getScriptFileNames,
            getScriptVersion,
            getScriptSnapshot,
            readDirectory: (path, extensions, exclude, include, depth) => {
                var _a, _b;
                const result = (_b = (_a = vueHost.readDirectory) === null || _a === void 0 ? void 0 : _a.call(vueHost, path, extensions, exclude, include, depth)) !== null && _b !== void 0 ? _b : [];
                for (const uri of sourceFiles.getUris()) {
                    const vuePath = shared.uriToFsPath(uri);
                    const vuePath2 = upath.join(path, upath.basename(vuePath));
                    if (upath.relative(path.toLowerCase(), vuePath.toLowerCase()).startsWith('..')) {
                        continue;
                    }
                    if (!depth && vuePath.toLowerCase() === vuePath2.toLowerCase()) {
                        result.push(vuePath2);
                    }
                    else if (depth) {
                        result.push(vuePath2); // TODO: depth num
                    }
                }
                return result;
            },
            getScriptKind(fileName) {
                switch (upath.extname(fileName)) {
                    case '.vue': return ts.ScriptKind.TSX; // can't use External, Unknown
                    case '.js': return ts.ScriptKind.JS;
                    case '.jsx': return ts.ScriptKind.JSX;
                    case '.ts': return ts.ScriptKind.TS;
                    case '.tsx': return ts.ScriptKind.TSX;
                    case '.json': return ts.ScriptKind.JSON;
                    default: return ts.ScriptKind.Unknown;
                }
            },
        };
        if (lsType === 'template') {
            tsHost.getCompilationSettings = () => ({
                ...vueHost.getCompilationSettings(),
                jsx: ts.JsxEmit.Preserve,
            });
        }
        return tsHost;
        function getScriptFileNames() {
            const tsFileNames = getLocalTypesFiles(lsType);
            for (const [tsUri] of sourceFiles.getTsDocuments(lsType)) {
                tsFileNames.push(shared.uriToFsPath(tsUri)); // virtual .ts
            }
            for (const fileName of vueHost.getScriptFileNames()) {
                if (isTsPlugin) {
                    tsFileNames.push(fileName); // .vue + .ts
                }
                else if (!fileName.endsWith('.vue')) {
                    tsFileNames.push(fileName); // .ts
                }
            }
            return tsFileNames;
        }
        function getScriptVersion(fileName) {
            const uri = shared.fsPathToUri(fileName);
            const basename = upath.basename(fileName);
            if (basename === localTypes.typesFileName) {
                return '0';
            }
            let doc = sourceFiles.getTsDocuments(lsType).get(uri);
            if (doc) {
                return doc.version.toString();
            }
            return vueHost.getScriptVersion(fileName);
        }
        function getScriptSnapshot(fileName) {
            const version = getScriptVersion(fileName);
            const cache = scriptSnapshots.get(fileName);
            if (cache && cache[0] === version) {
                return cache[1];
            }
            const basename = upath.basename(fileName);
            if (basename === localTypes.typesFileName) {
                return localTypesScript;
            }
            const uri = shared.fsPathToUri(fileName);
            const doc = sourceFiles.getTsDocuments(lsType).get(uri);
            if (doc) {
                const text = doc.getText();
                const snapshot = ts.ScriptSnapshot.fromString(text);
                scriptSnapshots.set(fileName, [version, snapshot]);
                return snapshot;
            }
            let tsScript = vueHost.getScriptSnapshot(fileName);
            if (tsScript) {
                if (lsType === 'template' && basename === 'runtime-dom.d.ts') {
                    // allow arbitrary attributes
                    let tsScriptText = tsScript.getText(0, tsScript.getLength());
                    tsScriptText = tsScriptText.replace('type ReservedProps = {', 'type ReservedProps = { [name: string]: any');
                    tsScript = ts.ScriptSnapshot.fromString(tsScriptText);
                }
                scriptSnapshots.set(fileName, [version, tsScript]);
                return tsScript;
            }
        }
    }
    function getHostDocument(uri) {
        const fileName = shared.uriToFsPath(uri);
        const version = Number(vueHost.getScriptVersion(fileName));
        if (!documents.uriHas(uri) || documents.uriGet(uri).version !== version) {
            const scriptSnapshot = vueHost.getScriptSnapshot(fileName);
            if (scriptSnapshot) {
                const scriptText = scriptSnapshot.getText(0, scriptSnapshot.getLength());
                const document = vscode_languageserver_textdocument_1.TextDocument.create(uri, uri.endsWith('.vue') ? 'vue' : 'typescript', version, scriptText);
                documents.uriSet(uri, document);
            }
        }
        if (documents.uriHas(uri)) {
            return documents.uriGet(uri);
        }
    }
    function updateSourceFiles(uris, shouldUpdateTemplateScript) {
        var _a;
        let vueScriptContentsUpdate = false;
        let vueScriptsUpdated = false;
        let templateScriptUpdated = false;
        if (shouldUpdateTemplateScript) {
            for (const cb of initProgressCallback) {
                cb(0);
            }
        }
        for (const uri of uris) {
            const sourceFile = sourceFiles.get(uri);
            const doc = getHostDocument(uri);
            if (!doc)
                continue;
            if (!sourceFile) {
                sourceFiles.set(uri, (0, sourceFile_1.createSourceFile)(doc.uri, doc.getText(), doc.version.toString(), context));
                vueScriptContentsUpdate = true;
                vueScriptsUpdated = true;
            }
            else {
                const updates = sourceFile.update(doc.getText(), doc.version.toString());
                if (updates.scriptContentUpdated) {
                    vueScriptContentsUpdate = true;
                }
                if (updates.scriptUpdated) {
                    vueScriptsUpdated = true;
                }
                if (updates.templateScriptUpdated) {
                    templateScriptUpdated = true;
                }
            }
            templateScriptUpdateUris.add(uri);
        }
        if (vueScriptContentsUpdate) {
            scriptContentVersion++;
        }
        if (vueScriptsUpdated) {
            scriptProjectVersion++;
            templateProjectVersion++;
        }
        if (shouldUpdateTemplateScript && lastScriptProjectVersionWhenTemplateProjectVersionUpdate !== scriptContentVersion) {
            lastScriptProjectVersionWhenTemplateProjectVersionUpdate = scriptContentVersion;
            let currentNums = 0;
            for (const uri of templateScriptUpdateUris) {
                if ((_a = sourceFiles.get(uri)) === null || _a === void 0 ? void 0 : _a.updateTemplateScript(templateTsLs)) {
                    templateScriptUpdated = true;
                }
                currentNums++;
                for (const cb of initProgressCallback) {
                    cb(currentNums / templateScriptUpdateUris.size);
                }
            }
            templateScriptUpdateUris.clear();
            for (const cb of initProgressCallback) {
                cb(1);
            }
            initProgressCallback.length = 0;
        }
        if (templateScriptUpdated) {
            templateProjectVersion++;
        }
    }
    function unsetSourceFiles(uris) {
        let updated = false;
        for (const uri of uris) {
            if (sourceFiles.delete(uri)) {
                updated = true;
            }
        }
        if (updated) {
            scriptContentVersion++;
            scriptProjectVersion++;
            templateProjectVersion++;
        }
    }
}
exports.createLanguageService = createLanguageService;
function createServices(ts, vueHost) {
    const fileSystemProvider = {
        stat: (uri) => {
            return new Promise((resolve, reject) => {
                fs.stat(shared.uriToFsPath(uri), (err, stats) => {
                    if (stats) {
                        resolve({
                            type: stats.isFile() ? html.FileType.File
                                : stats.isDirectory() ? html.FileType.Directory
                                    : stats.isSymbolicLink() ? html.FileType.SymbolicLink
                                        : html.FileType.Unknown,
                            ctime: stats.ctimeMs,
                            mtime: stats.mtimeMs,
                            size: stats.size,
                        });
                    }
                    else {
                        reject(err);
                    }
                });
            });
        },
        readDirectory: (uri) => {
            return new Promise((resolve, reject) => {
                fs.readdir(shared.uriToFsPath(uri), (err, files) => {
                    if (files) {
                        resolve(files.map(file => [file, html.FileType.File]));
                    }
                    else {
                        reject(err);
                    }
                });
            });
        },
    };
    const htmlLs = html.getLanguageService({ fileSystemProvider });
    const cssLs = css.getCSSLanguageService({ fileSystemProvider });
    const scssLs = css.getSCSSLanguageService({ fileSystemProvider });
    const lessLs = css.getLESSLanguageService({ fileSystemProvider });
    const pugLs = pug.getLanguageService(htmlLs);
    const jsonLs = json.getLanguageService({ schemaRequestService: vueHost === null || vueHost === void 0 ? void 0 : vueHost.schemaRequestService });
    const postcssLs = {
        ...scssLs,
        doValidation: (document, stylesheet, documentSettings) => {
            let errors = scssLs.doValidation(document, stylesheet, documentSettings);
            errors = errors.filter(error => error.code !== 'css-semicolonexpected');
            errors = errors.filter(error => error.code !== 'css-ruleorselectorexpected');
            errors = errors.filter(error => error.code !== 'unknownAtRules');
            return errors;
        },
    };
    let htmlDataProviders = [];
    return {
        ts,
        htmlLs,
        pugLs,
        jsonLs,
        getCssLs,
        vueHost,
        updateHtmlCustomData,
        updateCssCustomData,
        getHtmlDataProviders: () => htmlDataProviders,
    };
    function updateHtmlCustomData(customData) {
        htmlDataProviders = [];
        for (const id in customData) {
            htmlDataProviders.push(html.newHTMLDataProvider(id, customData[id]));
        }
        htmlLs.setDataProviders(true, htmlDataProviders);
    }
    function updateCssCustomData(customData) {
        const data = customData.map(data => css.newCSSDataProvider(data));
        cssLs.setDataProviders(true, data);
        scssLs.setDataProviders(true, data);
        lessLs.setDataProviders(true, data);
    }
    function getCssLs(lang) {
        switch (lang) {
            case 'css': return cssLs;
            case 'scss': return scssLs;
            case 'less': return lessLs;
            case 'postcss': return postcssLs;
        }
    }
}
//# sourceMappingURL=languageService.js.map