"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.register = exports.getSemanticTokenLegend = void 0;
const shared_1 = require("@vue/shared");
const ts2 = require("vscode-typescript-languageservice"); // TODO: remove it
function getSemanticTokenLegend() {
    const tsLegend = ts2.getSemanticTokenLegend();
    const tokenTypesLegend = [
        ...tsLegend.types,
        'componentTag',
        'operator', // namespaced component accessor: '.'
    ];
    const semanticTokenLegend = {
        tokenTypes: tokenTypesLegend,
        tokenModifiers: tsLegend.modifiers,
    };
    return semanticTokenLegend;
}
exports.getSemanticTokenLegend = getSemanticTokenLegend;
function register({ sourceFiles, getTsLs, htmlLs, pugLs, scriptTsLs, modules: { html } }, updateTemplateScripts) {
    const semanticTokensLegend = getSemanticTokenLegend();
    const tokenTypes = new Map(semanticTokensLegend.tokenTypes.map((t, i) => [t, i]));
    return (uri, range, cancle, reportProgress) => {
        const sourceFile = sourceFiles.get(uri);
        if (!sourceFile) {
            // take over mode
            const tokens = scriptTsLs.getDocumentSemanticTokens(uri, range, cancle);
            return tokens;
        }
        const document = sourceFile.getTextDocument();
        const offsetRange = range ?
            {
                start: document.offsetAt(range.start),
                end: document.offsetAt(range.end),
            } : {
            start: 0,
            end: document.getText().length,
        };
        let tokens = [];
        if (cancle === null || cancle === void 0 ? void 0 : cancle.isCancellationRequested)
            return;
        const scriptResult = getTsResult(sourceFile, 'script');
        if (!(cancle === null || cancle === void 0 ? void 0 : cancle.isCancellationRequested) && scriptResult.length) {
            tokens = tokens.concat(scriptResult);
            reportProgress === null || reportProgress === void 0 ? void 0 : reportProgress(tokens);
        }
        if (sourceFile.getHtmlSourceMaps().length) {
            updateTemplateScripts();
        }
        if (cancle === null || cancle === void 0 ? void 0 : cancle.isCancellationRequested)
            return;
        const templateResult = getTsResult(sourceFile, 'template');
        if (!(cancle === null || cancle === void 0 ? void 0 : cancle.isCancellationRequested) && templateResult.length) {
            tokens = tokens.concat(templateResult);
            reportProgress === null || reportProgress === void 0 ? void 0 : reportProgress(tokens);
        }
        if (cancle === null || cancle === void 0 ? void 0 : cancle.isCancellationRequested)
            return;
        const htmlResult = getHtmlResult(sourceFile);
        if (!(cancle === null || cancle === void 0 ? void 0 : cancle.isCancellationRequested) && htmlResult.length) {
            tokens = tokens.concat(htmlResult);
            reportProgress === null || reportProgress === void 0 ? void 0 : reportProgress(tokens);
        }
        return tokens;
        function getTsResult(sourceFile, lsType) {
            var _a;
            const result = [];
            for (const sourceMap of sourceFile.getTsSourceMaps()) {
                if (sourceMap.lsType !== lsType)
                    continue;
                const tsLs = getTsLs(sourceMap.lsType);
                for (const maped of sourceMap.mappings) {
                    if (maped.data.capabilities.semanticTokens
                        && maped.sourceRange.end > offsetRange.start
                        && maped.sourceRange.start < offsetRange.end) {
                        if (cancle === null || cancle === void 0 ? void 0 : cancle.isCancellationRequested)
                            return result;
                        const tsRange = {
                            start: sourceMap.mappedDocument.positionAt(maped.mappedRange.start),
                            end: sourceMap.mappedDocument.positionAt(maped.mappedRange.end),
                        };
                        const tokens = tsLs.getDocumentSemanticTokens(sourceMap.mappedDocument.uri, tsRange, cancle);
                        if (!tokens)
                            continue;
                        for (const token of tokens) {
                            const tsStart = sourceMap.mappedDocument.offsetAt({ line: token[0], character: token[1] });
                            const tsEnd = sourceMap.mappedDocument.offsetAt({ line: token[0], character: token[1] + token[2] });
                            const vueRange = (_a = sourceMap.getSourceRange(tsStart, tsEnd, data => !!data.capabilities.semanticTokens)) === null || _a === void 0 ? void 0 : _a[0];
                            if (!vueRange)
                                continue;
                            const vuePos = document.positionAt(vueRange.start);
                            result.push([vuePos.line, vuePos.character, vueRange.end - vueRange.start, token[3], token[4]]);
                        }
                    }
                }
            }
            return result;
        }
        function getHtmlResult(sourceFile) {
            var _a, _b, _c, _d;
            const result = [];
            const templateScriptData = sourceFile.getTemplateScriptData();
            const components = new Set([
                ...templateScriptData.components,
                ...templateScriptData.components.map(shared_1.hyphenate).filter(name => !(0, shared_1.isHTMLTag)(name)),
            ]);
            for (const sourceMap of [...sourceFile.getHtmlSourceMaps(), ...sourceFile.getPugSourceMaps()]) {
                let htmlStart = (_a = sourceMap.getMappedRange(offsetRange.start)) === null || _a === void 0 ? void 0 : _a[0].start;
                if (htmlStart === undefined) {
                    for (const mapping of sourceMap.mappings) {
                        if (mapping.sourceRange.end >= offsetRange.start) {
                            if (htmlStart === undefined || mapping.mappedRange.start < htmlStart) {
                                htmlStart = mapping.mappedRange.start;
                            }
                        }
                    }
                }
                if (htmlStart === undefined)
                    continue;
                const docText = sourceMap.mappedDocument.getText();
                const scanner = sourceMap.language === 'html'
                    ? htmlLs.createScanner(docText)
                    : pugLs.createScanner(sourceMap.pugDocument);
                if (!scanner)
                    continue;
                let token = scanner.scan();
                while (token !== html.TokenType.EOS) {
                    const tokenOffset = scanner.getTokenOffset();
                    if (tokenOffset >= htmlStart && (token === html.TokenType.StartTag || token === html.TokenType.EndTag)) {
                        const tokenText = scanner.getTokenText();
                        if (components.has(tokenText) || tokenText.indexOf('.') >= 0) {
                            const tokenLength = scanner.getTokenLength();
                            const vueRange = (_b = sourceMap.getSourceRange(tokenOffset)) === null || _b === void 0 ? void 0 : _b[0];
                            if (vueRange) {
                                const vueOffset = vueRange.start;
                                if (vueOffset > offsetRange.end)
                                    break; // TODO: fix source map perf and break in while condition
                                const vuePos = sourceMap.sourceDocument.positionAt(vueOffset);
                                if (components.has(tokenText)) {
                                    result.push([vuePos.line, vuePos.character, tokenLength, (_c = tokenTypes.get('componentTag')) !== null && _c !== void 0 ? _c : -1, undefined]);
                                }
                                else if (tokenText.indexOf('.') >= 0) {
                                    for (let i = 0; i < tokenText.length; i++) {
                                        if (tokenText[i] === '.') {
                                            result.push([vuePos.line, vuePos.character + i, 1, (_d = tokenTypes.get('operator')) !== null && _d !== void 0 ? _d : -1, undefined]);
                                        }
                                    }
                                }
                            }
                        }
                    }
                    token = scanner.scan();
                }
            }
            return result;
        }
    };
}
exports.register = register;
//# sourceMappingURL=semanticTokens.js.map